/**********************************************************************************************
 * 
 * 24 Channel Stimulator Firmware
 * Version 3.7 (X13 Hardware Revision)
 * 2018-05-06
 * Brian Neltner, Neltner Labs (neltnerb@neltnerlabs.com)
 * 
 * Description: After updating the firmware framework to use the serial receive library with the
 * X12 hardware, we made some small modifications to the hardware connections that result in
 * incompatible but functionally similar X13. This version of the firmware is for hardware target
 * X13 only.
 * 
 * Changelog:
 * Version 3.7 -- Modified so that external trigger also starts table.
 * 
 **********************************************************************************************/

#include <Arduino.h>
#include <vector>
#include <math.h>
#include <SPI.h>
#include "SerialEval.h"
#include <LiquidCrystal.h>
#include "AD9838.h"
#include "AD5543.h"
#include "AD7768.h"
#include "MAX14752.h"
#include "Multiplexer.h"
#include "stimulatorInterface.h"
#include "LED.h"

// Firmware defines.
#define title "Brain Stimulator Control"
#define FirmwareVersion "3.7"
#define BoardVersion "13"

// Definition for user-accessible pins.
#define PWR15VONpin CANRX
#define HVOFFpin 51
#define triggerPin 52
#define triggerMonitor 53

// Global setup for EEG syncing.
boolean EEGLockingFlag = false;
StimulatorInterface stimInterface;
boolean firstEEGCommand = false;
unsigned long startms = 0;

//void onDataFunction(float freq_, float amp_, float phase_){
//  EEGPhaseTime = micros();
//  EEGFreq = freq_;
//  EEGAmp = amp_;
//  EEGPhase = phase_;
//  EEGPhaseUpdated = true;
//}

// Set up a vector table for stimulation program storage.
unsigned long stimulationstepstarttime;
unsigned long stimulationstependtime;
unsigned long stimulationstepduration;

// Set up a flag to indicate whether we're currently running a program.
boolean stimulationrunning = false;

// This will always point to the currently running row of a stimulation table.
unsigned int stimulationrownum = 0;

struct stimulationrow {
  float preidletime; // The time since stimulation start that the row should be started.
  float rampuptime;
  float duration;
  float rampdowntime;
  float postidletime;
  float freq0;
  float amp0;
  float offset0;
  float freq1;
  float amp1;
  float offset1;
  unsigned int ch1load;
  unsigned int ch2load;
  unsigned int ch3load;
  unsigned int ch4load;
  unsigned int ch1pin;
  unsigned int ch2pin;
  unsigned int ch3pin;
  unsigned int ch4pin;
  unsigned int trigger;
  unsigned int chconfig;
};

std::vector<stimulationrow> stimulationtable;

// And this is our state machine for running a stimulation.
enum stimulationtablestate {
  WAIT,
  SETUP,
  PREIDLE,
  RAMPUP,
  STIMRUN,
  RAMPDOWN,
  POSTIDLE
} stimulationtablestate;

// Global variables.
boolean debugSerial = true; // Set to false to disable sending any responses to commands over the USB interface.
enum stimulationstate {RUN, SAFE, DEBUG} stimulationstate;

// Define and configure all peripherals and support classes.
LiquidCrystal lcd(28, 26, 25, 20, 21, 22); // Top LCD.
LiquidCrystal lcd2(28, 27, 25, 20, 21, 22); // Bottom LCD.
AD5543 MDAC0(37, 5, 2, lcd, 0); // CS Pin, 20mA Range Pin, Full-Scale Current for Scaling, LCD device, column to print.
AD5543 MDAC1(38, 5, 2, lcd, 20); // CS Pin, 20mA Range Pin, Full-Scale Current for Scaling, LCD device, column to print.
AD9838 DDS0(2684350, 39, 45, 42, 43, 46, lcd, 7); // MCLK Frequency, CS Pin, RST Pin, PHASE Pin, FRQ Pin, MSB Pin.
AD9838 DDS1(2684350, 40, 45, 44, 47, 49, lcd, 27); // MCLK Frequency, CS Pin, RST Pin, PHASE Pin, FRQ Pin, MSB Pin.
Multiplexer MUX(2, 3, lcd2); // Define multiplexer chain. LATCH on pin 2, CLR on pin 3. lcd2 is the correct display to use.
boolean MDAC0Ramping, MDAC1Ramping;
unsigned long MDAC0StartTime, MDAC1StartTime, MDAC0RampTime, MDAC1RampTime;
float MDAC0StartCurrent, MDAC1StartCurrent, MDAC0EndCurrent, MDAC1EndCurrent;
AD7768 ExtADC(12, 13, 30, 48, 19, 18, 17, 16, 23, 24, SDA1, SCL1); // CS, START, DCLK, DRDY, D0-D7 pins.
// ExtADC Note: 19 18 17 16 23 24 SDA1 SCL1 maps to real pins A10 A11 A12 A13 A14 A15 A17 A18 so that while it seems random they are in true order.
MAX14752 WSS0(31, 32, 33); // S0 pin, S1 pin, S2 pin.
MAX14752 WSS1(34, 35, 36); // S0 pin, S1 pin, S2 pin.
LED StatusLED(6, 7, 8);
LED StimulationLED(9, 10, 11);

// To handle the peculiar AD7768 data interface, we require a special interrupt dedicated to checking DCLK for rising edges.
// This special interrupt routine is to deal with the pecular data format of the AD7768. It is triggered whenever
// the ADC DRDY pulse falls, indicating the start of data. ADCData is an 8-width array of 32-bit integers that have been successfully received.
// ADCDataBuffer contains data being updated. ADCBitPointer is a byte indicating where in the buffer bits we are.

unsigned long ADCData[8];
volatile unsigned long ADCDataBuffer[32]; // PORTA goes from [0:31] for each bit array received serially.
volatile boolean ADCDataReady = false;

// These are the prototypes for each individual serial command handler.

serialcommand(getVersion);
serialcommand(stimulationAddRow);
serialcommand(stimulationTableClear);
serialcommand(stimulationTableRun);
serialcommand(setStimulationTableMode);
serialcommand(stimulationTableAbort);
serialcommand(EEGLock);
serialcommand(setModeRun);
serialcommand(setModeSafe);
serialcommand(setModeDebug);
serialcommand(setTriggerOn);
serialcommand(setTriggerOff);
serialcommand(stimulate);
serialcommand(adcRead);
serialcommand(readDiagnosticSignals);
serialcommand(dacWrite);
serialcommand(ddsWrite);
serialcommand(wssWrite);
serialcommand(setDCMode);
serialcommand(setACMode);
serialcommand(channelConfigAll);
serialcommand(channelConfigSingle);
serialcommand(channel1ConfigSingle);
serialcommand(channel2ConfigSingle);
serialcommand(channel3ConfigSingle);
serialcommand(channel4ConfigSingle);
serialcommand(channelPin);
serialcommand(channel1Electrode);
serialcommand(channel2Electrode);
serialcommand(channel3Electrode);
serialcommand(channel4Electrode);
serialcommand(setHighCurrentMode);
serialcommand(setMDAC0);
serialcommand(setMDAC1);
serialcommand(rampMDAC0);
serialcommand(rampMDAC1);
serialcommand(rampMDACs);
serialcommand(setDebug);

void setup() {
  StatusLED.begin();
  StimulationLED.begin();

  StatusLED.setColor(RED, 1);

  // Initialize the LCD.
  lcd.begin(40, 2);
  lcd2.begin(40, 2);
  lcd.clear();
  lcd2.clear();
  lcd2.print("");
  lcd.setCursor(0, 0);
  lcd.print(String(title) + ": Disconnected");
  lcd.setCursor(0, 1);
  lcd.print("Firmware Version: " + String(FirmwareVersion));
  lcd.setCursor(0, 0);
  lcd2.print("Hardware Version: " + String(BoardVersion));
  lcd2.setCursor(0, 1);
  lcd2.print("Starting System.");
  lcd2.display();
  lcd2.display();
  
  delay(5000);
  
  SerialUSB.begin(9600); // Initialize the USB interface.
  SPI.begin();
  stimInterface.setDataHandler(&evaluateEEGcommand);
  stimInterface.begin(&Serial1, 115200); // Changed to Serial1 interface.
  SerialUSB.println("Starting System");

  // Enable the +/- 15V supplies.
  pinMode(PWR15VONpin, OUTPUT);
  digitalWrite(PWR15VONpin, HIGH);

  pinMode(HVOFFpin, OUTPUT);
  digitalWrite(HVOFFpin, LOW);

  // Configure onboard ADCs.
  analogReadResolution(16); // 16 bit resolution
  analogReference(AR_DEFAULT); // 3.3 Vref

  // Configure onboard DAC.
  analogWriteResolution(12); // 12 bit resolution is the maximum available.
  pinMode(DAC0, OUTPUT);
  pinMode(DAC1, OUTPUT);
  analogWrite(DAC0, 2048); // Set these to midscale by default for no offset.
  analogWrite(DAC1, 2048);

  // Initialize the peripherals.
  MUX.begin();
  DDS0.begin(10.0); // Argument is starting default frequency.
  DDS1.begin(10.0);
  MDAC0.begin(0); // Argument is starting default current.
  MDAC1.begin(0);
  MDAC0Ramping = false; // Initialize the state of the MDAC to not-ramping.
  MDAC1Ramping = false;
  ExtADC.begin(); // Start the AD7768.
  WSS0.begin(); // Start the waveform source selector.
  WSS1.begin();

  // Configure the trigger pins.
  pinMode(triggerPin, OUTPUT);
  pinMode(triggerMonitor, INPUT);
  digitalWrite(triggerPin, LOW); // Turn off the trigger output.
  attachInterrupt(digitalPinToInterrupt(triggerMonitor), triggerExtIntHandler, RISING);

  // Configure the ADC peculiar data interface interrupt.
  attachInterrupt(digitalPinToInterrupt(ExtADC.getDataPins()[1]), ADCDataHandler, FALLING);

  // Set the initial state of the system to RUN.
  stimulationstate = RUN;

  // Activate the MUX switches for run mode and allow MDAC to output currents.
  MUX.Enable(); // This is for the CLEAR line.
  MUX.EnableOutput(); // This tells it that it may connect to the electrodes.
  MUX.ACMode(); // Use AC coupling by default, software can change.
  MDAC0.Enable();
  MDAC1.Enable();

  // Register serial commands to the serial library.
  registerCommand("GetVersion", getVersion, "Get the current firmware version.");
  registerCommand("StimulationAddRow", stimulationAddRow, "Add a row to the stimulation table.");
  registerCommand("StimulationTableClear", stimulationTableClear, "Clear the current stimulation table.");
  registerCommand("StimulationTableRun", stimulationTableRun, "Start running the uploaded stimulation program.");
  registerCommand("StimulationTableMode", setStimulationTableMode, "Set Stimulation Table Mode (0 = Normal, 1 = External Trigger).");
  registerCommand("StimulationTableAbort", stimulationTableAbort, "Abort the currently running stimulation program.");
  registerCommand("EEGLock", EEGLock, "Engage EEG Locking.");
  registerCommand("SetModeRun", setModeRun, "Run mode, current on, MUX connected as programmed.");
  registerCommand("SetModeSafe", setModeSafe, "Current off, MUX connected to 10k loads only (no electrode connection).");
  registerCommand("SetModeDebug", setModeDebug, "Current on, MUX connected to 10k loads only (no electrode connection).");
  registerCommand("SetTriggerOn", setTriggerOn, "Toggle Trigger output pin high.");
  registerCommand("SetTriggerOff", setTriggerOff, "Toggle Trigger output pin low.");
  registerCommand("Stimulate", stimulate, "Start Stimulation. Expects stimulation period in milliseconds.");
  registerCommand("ADCRead", adcRead, "Sample 4000 ADC values from monitors.");
  registerCommand("ReadDiagnosticSignals", readDiagnosticSignals, "Read voltage monitors.");
  registerCommand("DAC", dacWrite, "Set DAC output value.");
  registerCommand("DDS", ddsWrite, "Configure digital synthesizer chip. Expects Channel number and frequency.");
  registerCommand("WSS", wssWrite, "Control waveform source selection.");
  registerCommand("SetACMode", setACMode, "Set Waveform Generation to AC Coupled.");
  registerCommand("SetDCMode", setDCMode, "Set Waveform Generation to DC Coupled.");
  registerCommand("ChConfig", channelConfigAll, "Configure the electrode output connections.");
  registerCommand("ChLoad", channelConfigSingle, "Configure a given output with a provided load code.");
  registerCommand("Ch1Load", channel1ConfigSingle, "Same as 'ChLoad 1'.");
  registerCommand("Ch2Load", channel2ConfigSingle, "Same as 'ChLoad 2'.");
  registerCommand("Ch3Load", channel3ConfigSingle, "Same as 'ChLoad 3'.");
  registerCommand("Ch4Load", channel4ConfigSingle, "Same as 'ChLoad 4'.");
  registerCommand("ChPin", channelPin, "Set channel electrode connection.");
  registerCommand("Ch1Electrode", channel1Electrode, "Same as 'ChPin 1'.");
  registerCommand("Ch2Electrode", channel2Electrode, "Same as 'ChPin 2'.");
  registerCommand("Ch3Electrode", channel3Electrode, "Same as 'ChPin 3'.");
  registerCommand("Ch4Electrode", channel4Electrode, "Same as 'ChPin 4'.");
  registerCommand("SetHighCurrentMode", setHighCurrentMode, "Set to 0-20mA range vs 0-2.0mA range (0/1).");
  registerCommand("Wfm0", setMDAC0, "Wfm0 mA <value>. Set the scaling on MDAC0.");
  registerCommand("Wfm1", setMDAC1, "Wfm1 mA <value>. Set the scaling on MDAC1.");
  registerCommand("RampWfm0", rampMDAC0, "RampWfm0 mA <startCurrent> <endCurrent> <period (ms)>. Ramp output scaling for MDAC0.");
  registerCommand("RampWfm1", rampMDAC1, "RampWfm0 mA <startCurrent> <endCurrent> <period (ms)>. Ramp output scaling for MDAC0.");
  registerCommand("RampWfm", rampMDACs, "RampWfm mA <startCurrent0> <endCurrent0> <startCurrent1> <endCurrent1> <period (ms)>. Simultaneous MDAC ramp.");
  registerCommand("SetDebug", setDebug, "Change USB Communication to silent or verbose responses.");

  ExtADC.PulseStart();
  
  lcd.setCursor(0, 1);
  lcd.print("0.00mA 10.00Hz      0.00mA 10.00Hz      ");
  
  StatusLED.setColor(GREEN, 1);
}

boolean calibrated = false;
unsigned int datacount = 0;
boolean serialConnected = false;
unsigned int stimulationTableMode = 0;

void loop() {
  // Check for an active USB connection.
  if(!serialConnected and SerialUSB) {
    StatusLED.setColor(BLUE, 1);
    lcd.setCursor(26, 0);
    lcd.print("Connected     ");
    serialConnected = true;
  }
  else if (serialConnected and !SerialUSB) {
    StatusLED.setColor(GREEN, 1);
    lcd.setCursor(26, 0);
    lcd.print("Disconnected  ");
    serialConnected = false;
  }

  checkSerial(); // Check for any new received commands on the USB interface.

  // ADC Data is stored presented in D0-D7 order (per bit of the 32-bit value) on A10 A11 A12 A13 A14 A15 A17 and A18, with bitshifting below.
  if (ADCDataReady) {
    ADCDataReady = false; // Clear the conversion flag.

    // Put the raw ADC data into per-channel format.
    ADCData[0] = 0;
    ADCData[1] = 0;
    ADCData[2] = 0;
    ADCData[3] = 0;
    ADCData[4] = 0;
    ADCData[5] = 0;
    ADCData[6] = 0;
    ADCData[7] = 0;
    
    for (byte i=0; i<32; i++) {
      ADCData[0] |= ((ADCDataBuffer[i] & (1<<10))>>10)<<i;
      ADCData[1] |= ((ADCDataBuffer[i] & (1<<11))>>11)<<i;
      ADCData[2] |= ((ADCDataBuffer[i] & (1<<12))>>12)<<i;
      ADCData[3] |= ((ADCDataBuffer[i] & (1<<13))>>13)<<i;
      ADCData[4] |= ((ADCDataBuffer[i] & (1<<14))>>14)<<i;
      ADCData[5] |= ((ADCDataBuffer[i] & (1<<15))>>15)<<i;
      ADCData[6] |= ((ADCDataBuffer[i] & (1<<17))>>17)<<i;
      ADCData[7] |= ((ADCDataBuffer[i] & (1<<18))>>18)<<i;
    }
    
    // Send these values to the ADC object for conversion to actual voltages.
    ExtADC.writeValue(ADCData); // Convert and store the raw data that has been received.

    if (!calibrated) {
      if (datacount >= 100) {
        ExtADC.calibrate();
        calibrated = true;
      }
      datacount++;
    }
  }

  // Record a stable current microsecond time.
  unsigned long currentTime = micros();

  // Handle any MDAC ramping curves.

  // Code to handle a stimulation table.
  if (stimulationrunning) { // This is the flag indicating that stimulation is active.
    switch (stimulationtablestate) {
      case WAIT:
        break;
      case SETUP:
        StimulationLED.setColor(120, 1);
        DDS0.SetFreq(stimulationtable[stimulationrownum].freq0); // Set DDS0 frequency.
        DDS1.SetFreq(stimulationtable[stimulationrownum].freq1); // Set DDS1 frequency.
        MDAC0.SetCurrent(0); // Set current 0 output to zero for start.
        MDAC1.SetCurrent(0); // Set current 1 output to zero for start.
        WSS0.setSource(SRC_SINE0); // Set waveform 0 to sin0.
        WSS1.setSource(SRC_SINE1); // Set waveform 1 to sin1.
        stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].preidletime * 1000000; // Add the preidletime in milliseconds to the current time.
        stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
        stimulationtablestate = PREIDLE; // And with the setup done we move to the PREIDLE stage.
        break;
      case PREIDLE:
        if (currentTime >= stimulationstependtime) { // If the correct number of milliseconds have passed...
          MUX.setElectrode(1, stimulationtable[stimulationrownum].ch1pin); // Configure the electrode output pins.
          MUX.setElectrode(2, stimulationtable[stimulationrownum].ch2pin);
          MUX.setElectrode(3, stimulationtable[stimulationrownum].ch3pin);
          MUX.setElectrode(4, stimulationtable[stimulationrownum].ch4pin);
          MUX.setLoad(1, stimulationtable[stimulationrownum].ch1load); // Configure the loads.
          MUX.setLoad(2, stimulationtable[stimulationrownum].ch2load);
          MUX.setLoad(3, stimulationtable[stimulationrownum].ch3load);
          MUX.setLoad(4, stimulationtable[stimulationrownum].ch4load);
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].rampuptime * 1000000; // Add the rampup time in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          MDAC0.SetScaler(0);
          MDAC1.SetScaler(0);
          MDAC0.SetCurrent(stimulationtable[stimulationrownum].amp0);
          MDAC1.SetCurrent(stimulationtable[stimulationrownum].amp1);
          stimulationtablestate = RAMPUP; // And with the preidle step done we move to the RAMPUP stage.
          digitalWrite(triggerPin, stimulationtable[stimulationrownum].trigger); // Send the trigger output for this row.
        }
        break;
      case RAMPUP:
        if (currentTime >= stimulationstependtime) { // Move on if the ramp is over.
          MDAC0.SetScaler(1);
          MDAC1.SetScaler(1);
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].duration * 1000000; // Add the duration in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          stimulationtablestate = STIMRUN; // And with the ramp up done we move to the STIMRUN stage.
          StimulationLED.setColor(240, 1);
        }
        else { // Otherwise, set to ramp output up, non-blocking.
          float scaler = (1 / (float)stimulationstepduration) * (float)(currentTime - stimulationstepstarttime);
          MDAC0.SetScaler(scaler);
          MDAC1.SetScaler(scaler);
          StimulationLED.setColor(120+(float)120*scaler, 1);
        }
        break;
      case STIMRUN:
        if (currentTime >= stimulationstependtime) { // Move on if the stimulation pulse is over.
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].rampdowntime * 1000000; // Add the duration in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          stimulationtablestate = RAMPDOWN; // And with the stimulation pulse done we move to the RAMPDOWN stage.
        }
        break;
      case RAMPDOWN:
        if (currentTime >= stimulationstependtime) { // Move on if the ramp down is over.
          MDAC0.SetCurrent(0); // Set both current outputs to zero, otherwise leave the configuration as-is.
          MDAC1.SetCurrent(0);
          MDAC0.SetScaler(1);
          MDAC1.SetScaler(1);
          if (stimulationtable[stimulationrownum].chconfig == 1) { // If the Channel Config is one, reset the channels to a safe mode after ramp down.
            MUX.setElectrode(1, 0); // Configure the electrode output pins to disconnected.
            MUX.setElectrode(2, 0);
            MUX.setElectrode(3, 0);
            MUX.setElectrode(4, 0);
            MUX.setLoad(1, 8); // Configure the loads to the 1k resistor.
            MUX.setLoad(2, 8);
            MUX.setLoad(3, 8);
            MUX.setLoad(4, 8);
          }
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].postidletime * 1000000; // Add the duration in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          digitalWrite(triggerPin, LOW); // Reset the trigger output for this row.
          stimulationtablestate = POSTIDLE; // And with the ramp down done we move to the POSTIDLE stage.
          StimulationLED.setColor(120, 1);
        }
        else { // Otherwise, set to ramp output down, non-blocking.
          float scaler = (1 / (float)stimulationstepduration) * (float)(currentTime - stimulationstepstarttime);
          MDAC0.SetScaler(1-scaler);
          MDAC1.SetScaler(1-scaler);
          StimulationLED.setColor(240-(float)120*scaler, 1);
        }
        break;
      case POSTIDLE:
        if (currentTime >= stimulationstependtime) { // If the postidle time is done, increment the row number and go back to SETUP.
          if (stimulationTableMode != 1) {
            stimulationrownum++; // Increment the row number.
            if (stimulationrownum >= stimulationtable.size()) {
              stimulationrunning = false; // If the table is all done, turn off stimulation.
              StimulationLED.setColor(GREEN, 0);
            }
            else {
              stimulationtablestate = SETUP; // Otherwise, move to the next line's SETUP state. 
              stimulationstepstarttime = stimulationstependtime;
            }
          }
        }
        break;
      default:
        stimulationrunning = false;
        MDAC0.SetCurrent(0); // Turn off both output DACs.
        MDAC1.SetCurrent(0);
        MUX.setLoad(1, 8); // Connect Channel Loads to the 1k resistor, disconnect from electrodes.
        MUX.setLoad(2, 8);
        MUX.setLoad(3, 8);
        MUX.setLoad(4, 8);
        MUX.setElectrode(1, 0);
        MUX.setElectrode(2, 0);
        MUX.setElectrode(3, 0);
        MUX.setElectrode(4, 0);
        StimulationLED.setColor(RED, 1);
        SerialUSB.println("ERROR CODE 329157: Stimulation Table Execution has a firmware bug assigning invalid state. Stimulation aborted and set to safe mode.");
        break;
    }
  }
  
  if (MDAC0Ramping) {
    unsigned long MDAC0RampTimer = currentTime - MDAC0StartTime;
    if (MDAC0RampTimer >= MDAC0RampTime) {
      MDAC0Ramping = false;
      MDAC0.SetCurrent(MDAC0EndCurrent);
    }
    else MDAC0.SetCurrent(MDAC0StartCurrent + (MDAC0EndCurrent-MDAC0StartCurrent)*((float)MDAC0RampTimer/(float)MDAC0RampTime));
  }

  if (MDAC1Ramping) {
    unsigned long MDAC1RampTimer = currentTime - MDAC1StartTime;
    if (MDAC1RampTimer >= MDAC1RampTime) {
      MDAC1Ramping = false;
      MDAC1.SetCurrent(MDAC1EndCurrent);
    }
    else MDAC1.SetCurrent(MDAC1StartCurrent + (MDAC1EndCurrent-MDAC1StartCurrent)*((float)MDAC1RampTimer/(float)MDAC1RampTime));
  }
}

// Special extra function for EEG data conversion. (returns current EEGpos-EEGneg in millivolts)
float getEEG(void) {
  return ExtADC.getValue(0)*1000;
}

void triggerExtIntHandler() {
  if (stimulationTableMode == 1) { // Ignore triggers in automatic advance mode.
    if (stimulationrunning) {
      stimulationrownum++; // Increment the row number.
      if (stimulationrownum >= stimulationtable.size()) {
        stimulationrunning = false; // If the table is all done, turn off stimulation.
        StimulationLED.setColor(GREEN, 0);
      }
      else {
        stimulationtablestate = SETUP; // Otherwise, move to the next line's SETUP state. 
        stimulationstepstarttime = micros();
      }
    }
  }
}

void ADCDataHandler() {
  // Entering this handler implies a falling edge of DRDY. Once that happens, the DCLK is so fast that there is no choice but to handle it with unrolled polling.
  // REG_PIOx_PDSR is the port reading register connected to physical SAM3X pins. The syntax (REG_PIOD_PDSR) & (1<<9), for instance, is zero if pin C9 is low
  // (arduino pin 48 used as the clock input) and 0b1000000000 if pin C9 is high. In this I simply copy the whole port into the data buffer for postprocessing due
  // to speed limitations.

  // ExtADC Note: 19 18 17 16 23 24 SDA1 SCL1 (D0-7 in order) maps to real pins A10 A11 A12 A13 A14 A15 A17 A18 so that while it seems random they are in true order.
  // With this exact wire connection I can execute this read in one line instead of reading all three ports for every bit.

  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[31] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[30] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[29] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[28] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[27] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[26] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[25] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[24] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[23] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[22] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[21] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[20] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[19] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[18] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[17] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[16] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[15] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[14] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[13] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[12] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[11] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[10] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[9] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[8] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[7] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[6] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[5] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[4] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[3] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[2] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[1] = REG_PIOA_PDSR;
  while(!(REG_PIOD_PDSR & (1<<9)));
  while((REG_PIOD_PDSR) & (1<<9));
  ADCDataBuffer[0] = REG_PIOA_PDSR;
  
  ADCDataReady = true;
}

void evaluateEEGcommand (float freq_, float amp_, float phase_) {
  if(firstEEGCommand){
    firstEEGCommand = false;
    startms = millis();
  }
  
  unsigned long elapsedms = millis()-startms;
  double elapseds = elapsedms/1000.0;
  double phase = phase_+PI; // 2*PI*freq_*elapseds + 
  double unity_wave = 0.5+cos(phase)/2.0; // the amplitude ranges from 0 to 1

  MDAC0.SetScaler(unity_wave);
  MDAC1.SetScaler(unity_wave);
}

/*******************************************************************************************************
 * 
 * Command Handlers. Generally USB responses are only sent if debugSerial is true. The exception is
 * the SetDebug command, which always responds since if they're using it they probably need that info.
 * 
 * Every single distinct serial command has an associated handler below which is automatically called
 * if the serial command is received. In each case, numArgs is used to guarantee that the general
 * command format is valid, the intArg and floatArg functions pull in arguments from the serial command
 * while doing typechecking (it returns with an error if the data types are mismatched), and data
 * validation is immediately performed to test that the arguments are valid for our command.
 * 
 * Some commands are goofy, for legacy reasons. They are emulating the X2 firmware (version Stim24ChX72)
 * to maintain compatibility with old control scripts and old versions of the LabView software. New
 * commands are introduced to allow access to new functionality not present on the X2 board.
 * 
 *******************************************************************************************************/

serialcommand(getVersion) {
  SerialUSB.println(String(title) + " Version: " + String(FirmwareVersion));
}

serialcommand(EEGLock) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: EEGLock expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int enableEEG = intArg(1);
  if ((enableEEG != 0) and (enableEEG != 1)) {
    if (debugSerial) SerialUSB.println("ERROR: EEGLock expects either a 0 or a 1 to disable/enable EEG Locking.");
    return;
  }

  if (enableEEG == 0) {
    EEGLockingFlag = false;
    if (debugSerial) SerialUSB.println("SUCCESS: EEG Locking Disabled.");
  }
  else if (enableEEG == 1) {
    EEGLockingFlag = true;
    firstEEGCommand = true;
    DDS0.Synchronize();
    DDS1.Synchronize();
    if (debugSerial) SerialUSB.println("SUCCESS: EEG Locking Enabled.");
  }
}

serialcommand(stimulationAddRow) {
  if (numArgs() != 21) {
    if (debugSerial) SerialUSB.println("ERROR: StimulationAddRow expects 21 arguments, received " + String(numArgs()) + ".");
    return;
  }
  
  stimulationrow newrow;
  newrow.preidletime = floatArg(1);
  newrow.rampuptime = floatArg(2);
  newrow.duration = floatArg(3);
  newrow.rampdowntime = floatArg(4);
  newrow.postidletime = floatArg(5);
  newrow.freq0 = floatArg(6);
  newrow.amp0 = floatArg(7);
  newrow.offset0 = floatArg(8);
  newrow.ch1load = intArg(9);
  newrow.ch1pin = intArg(10);
  newrow.ch2load = intArg(11);
  newrow.ch2pin = intArg(12);
  newrow.freq1 = floatArg(13);
  newrow.amp1 = floatArg(14);
  newrow.offset1 = floatArg(15);
  newrow.ch3load = intArg(16);
  newrow.ch3pin = intArg(17);
  newrow.ch4load = intArg(18);
  newrow.ch4pin = intArg(19);
  newrow.trigger = intArg(20);
  newrow.chconfig = intArg(21);

  if (newrow.preidletime < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Pre-Idle Time must be positive.");
    return;
  }
  if (newrow.rampuptime < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Ramp Up Time must be positive.");
    return;
  }
  if (newrow.duration < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Duration must be positive.");
    return;
  }
  if (newrow.rampdowntime < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Ramp Down Time must be positive.");
    return;
  }
  if (newrow.postidletime < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Post-Idle Time must be positive.");
    return;
  }
  if (newrow.freq0 <= 0 or newrow.freq0 > 350000) {
    if (debugSerial) SerialUSB.println("ERROR: DDS0 Frequency must be greater than zero and less than 350000Hz.");
    return;
  }
  if (newrow.freq1 <= 0 or newrow.freq1 > 350000) {
    if (debugSerial) SerialUSB.println("ERROR: DDS1 Frequency must be greater than zero and less than 350000Hz.");
    return;
  }
  if (newrow.amp0 < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Amplitude 0 must be positive.");
    return;
  }
  if (newrow.amp1 < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Amplitude 1 must be positive.");
    return;
  }
  if (newrow.offset0 < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Offset 0 must be positive.");
    return;
  }
  if (newrow.offset1 < 0) {
    if (debugSerial) SerialUSB.println("ERROR: Offset 1 must be positive.");
    return;
  }
  if (newrow.ch1load > 255) {
    if (debugSerial) SerialUSB.println("ERROR: Ch1Load is invalid.");
    return;
  }
  if (newrow.ch2load > 255) {
    if (debugSerial) SerialUSB.println("ERROR: Ch2Load is invalid.");
    return;
  }
  if (newrow.ch3load > 255) {
    if (debugSerial) SerialUSB.println("ERROR: Ch3Load is invalid.");
    return;
  }
  if (newrow.ch4load > 255) {
    if (debugSerial) SerialUSB.println("ERROR: Ch4Load is invalid.");
    return;
  }
  if (newrow.ch1pin > 24) {
    if (debugSerial) SerialUSB.println("ERROR: Ch1 Electrode is invalid.");
    return;
  }
  if (newrow.ch2pin > 24) {
    if (debugSerial) SerialUSB.println("ERROR: Ch2 Electrode is invalid.");
    return;
  }
  if (newrow.ch3pin > 24) {
    if (debugSerial) SerialUSB.println("ERROR: Ch3 Electrode is invalid.");
    return;
  }
  if (newrow.ch4pin > 24) {
    if (debugSerial) SerialUSB.println("ERROR: Ch4 Electrode is invalid.");
    return;
  }
  if (newrow.trigger != 0 and newrow.trigger != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Trigger Configuration must be zero or one.");
    return;    
  }
  if (newrow.chconfig != 0 and newrow.chconfig != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Channel Configuration (whether electrodes are reset after rampdown) must be zero or one.");
    return;    
  }
  stimulationtable.push_back(newrow);
  if (debugSerial) SerialUSB.println("SUCCESS: Added new row to the stimulation table.");
}

serialcommand(stimulationTableClear) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: StimulationTableClear expects zero arguments, received " + String(numArgs()) + ".");
    return;
  }
  stimulationtable.clear();
  if (debugSerial) SerialUSB.println("SUCCESS: Cleared stimulation table.");
}

serialcommand(stimulationTableRun) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: StimulationTableRun expects zero arguments, received " + String(numArgs()) + ".");
    return;
  }
  if (stimulationtable.size() > 0) {
    stimulationstepstarttime = micros();
    stimulationrownum = 0;
    if (stimulationTableMode != 1) {
      stimulationrunning = true;
      stimulationtablestate = SETUP;
      if (debugSerial) SerialUSB.println("SUCCESS: Started stimulation table.");
    }
    else if (stimulationTableMode == 1) {
      stimulationrunning = true;
      stimulationtablestate = WAIT;
      StimulationLED.setColor(60, 1);
      MDAC0.SetCurrent(0); // Set current 0 output to zero for start.
      MDAC1.SetCurrent(0); // Set current 1 output to zero for start.
      if (debugSerial) SerialUSB.println("SUCCESS: Stimulation table queued and awaiting start trigger.");
    }
  }
  else {
    if (debugSerial) SerialUSB.println("ERROR: Stimulation table is empty, try uploading some rows first.");
  }
}

serialcommand(setStimulationTableMode) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: StimulationTableMode expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int setting = intArg(1);
  if (setting == 0) {
    stimulationTableMode = 0;
    if (debugSerial) SerialUSB.println("SUCCESS: Stimulation Table Mode set to progress automatically between rows.");
  }
  else if (setting == 1) {
    stimulationTableMode = 1;
    if (debugSerial) SerialUSB.println("SUCCESS: Stimulation Table Mode set to progress between rows based on external trigger monitor.");
  }
  else {
    if (debugSerial) SerialUSB.println("ERROR: Stimulation Table Mode must be either 0 (normal) or 1 (external trigger).");
  }
}

serialcommand(stimulationTableAbort) {
  // No argument checking because this is a critical function.
  stimulationrunning = false;
  MDAC0.SetCurrent(0);
  MDAC1.SetCurrent(0);
  MUX.setElectrode(1, 0);
  MUX.setElectrode(2, 0);
  MUX.setElectrode(3, 0);
  MUX.setElectrode(4, 0);
  MUX.setLoad(1, 8); // Connect to 1k resistor.
  MUX.setLoad(2, 8);
  MUX.setLoad(3, 8);
  MUX.setLoad(4, 8);
  StimulationLED.setColor(GREEN, 0);
  if (debugSerial) SerialUSB.println("SUCCESS: Disconnected all electrodes, set all loads to 1k, set current to zero, stopped stimulation table.");
}

serialcommand(setDebug) {
  if (numArgs() != 1) {
    SerialUSB.println("ERROR: SetDebug expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int setting = intArg(1);

  if ((setting == 0) or (setting == 1)) {
    debugSerial = setting;
    SerialUSB.println("SUCCESS: USB Serial Debug Information set to " + String(setting) + ".");
  }
  else SerialUSB.println("ERROR: SetDebug must be set to either 0 (off) or 1 (on).");
}

serialcommand(setTriggerOn) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetTriggerOn expects zero arguments, received " + String(numArgs()) + ".");
    return;
  }

  digitalWrite(triggerPin, HIGH);
  if (debugSerial) SerialUSB.println("SUCCESS: Trigger Pin set to HIGH.");
}

serialcommand(setTriggerOff) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetTriggerOff expects zero arguments, received " + String(numArgs()) + ".");
    return;
  }

  digitalWrite(triggerPin, LOW);
  if (debugSerial) SerialUSB.println("SUCCESS: Trigger Pin set to LOW.");
}

serialcommand(ddsWrite) {
  if (numArgs() != 2) {
    if (debugSerial) SerialUSB.println("ERROR: DDS expects two arguments, received " + String(numArgs()) + ".");
    return;
  }

  int channel = intArg(1);  
  float frequency = floatArg(2);
  
  if ((frequency >= 0) and (frequency <= DDS0.GetMCLK())) {
    if (channel == 0) {
      DDS0.SetFreq(frequency);
      DDS0.Synchronize();
    }
    else if (channel == 1) {
      DDS1.SetFreq(frequency);
      DDS1.Synchronize();
    }
    else if (debugSerial) SerialUSB.println("ERROR: Channel must be 0 or 1.");
        
    if (debugSerial) SerialUSB.println("SUCCESS: DDS " + String(channel) + " set to " + String(frequency) + " Hz.");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Frequency must be between zero and " + String(DDS0.GetMCLK()) + " Hz.");
}

serialcommand(channelConfigAll) {
  if (numArgs() != 8) {
    if (debugSerial) SerialUSB.println("ERROR: ChConfig expects eight arguments, received " + String(numArgs()) + ".");
    return;
  }

  int load1 = intArg(1);
  int load2 = intArg(2);
  int load3 = intArg(3);
  int load4 = intArg(4);
  int electrode1 = intArg(5);
  int electrode2 = intArg(6);
  int electrode3 = intArg(7);
  int electrode4 = intArg(8);

  // Data validation.

  if ((load1 >= 0) and (load1 <= 0xFF) and (load2 >= 0) and (load2 <= 0xFF) and (load3 >= 0) and (load3 <= 0xFF) and (load4 >= 0) and (load4 <= 0xFF) and (electrode1 >= 0) and (electrode1 <= 24) and (electrode2 >= 0) and (electrode2 <= 24) and (electrode3 >= 0) and (electrode3 <= 24) and (electrode4 >= 0) and (electrode4 <= 24)) {
    MUX.setElectrode(1, electrode1);
    MUX.setElectrode(2, electrode2);
    MUX.setElectrode(3, electrode3);
    MUX.setElectrode(4, electrode4);
    MUX.setLoad(1, load1);
    MUX.setLoad(2, load2);
    MUX.setLoad(3, load3);
    MUX.setLoad(4, load4);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Loads and Electrode Configuration Set.");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Invalid loads and or electrode values. Loads must be from 0-255 and Electrodes must be from 0-24.");
}


serialcommand(channelConfigSingle) {
  if (numArgs() != 2) {
    if (debugSerial) SerialUSB.println("ERROR: ChLoad expects two arguments, received " + String(numArgs()) + ".");
    return;
  }

  int channel = intArg(1);
  int load = intArg(2);

  // Data Validation
  if ((channel >= 1) and (channel <= 4) and (load >= 0) and (load <= 0xFF)) {
    MUX.setLoad(channel, load);
    
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel " + String(channel) + " load set to " + String(load) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Channel must be between 1 and 4, and load must be between 0 and 255.");
}

serialcommand(channel1ConfigSingle) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: ChLoad1 expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int load = intArg(1);

  // Data Validation
  if ((load >= 0) and (load <= 0xFF)) {
    MUX.setLoad(1, load);
    
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 1 load set to " + String(load) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Load must be between 0 and 255.");
}

serialcommand(channel2ConfigSingle) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: ChLoad2 expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int load = intArg(1);

  // Data Validation
  if ((load >= 0) and (load <= 0xFF)) {
    MUX.setLoad(2, load);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 2 load set to " + String(load) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Load must be between 0 and 255.");
}

serialcommand(channel3ConfigSingle) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: ChLoad3 expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int load = intArg(1);

  // Data Validation
  if ((load >= 0) and (load <= 0xFF)) {
    MUX.setLoad(3, load);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 3 load set to " + String(load) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Load must be between 0 and 255.");
}

serialcommand(channel4ConfigSingle) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: ChLoad4 expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int load = intArg(1);

  // Data Validation
  if ((load >= 0) and (load <= 0xFF)) {
    MUX.setLoad(4, load);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 4 load set to " + String(load) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Load must be between 0 and 255.");
}

serialcommand(channelPin) {
  if (numArgs() != 2) {
    if (debugSerial) SerialUSB.println("ERROR: ChPin expects two arguments, received " + String(numArgs()) + ".");
    return;
  }

  int channel = intArg(1);
  int electrode = intArg(2);

  // Data Validation
  if ((channel >= 1) and (channel <= 4) and (electrode >= 0) and (electrode <= 24)) {
    MUX.setElectrode(channel, electrode);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel " + String(channel) + " electrode set to " + String(electrode) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Channel must be between 1 and 4, and electrode must be between 0 and 24.  0 indicates disconnected.");
}

serialcommand(channel1Electrode) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Ch1Electrode expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int electrode = intArg(1);

  // Data Validation
  if ((electrode >= 0) and (electrode <= 24)) {
    MUX.setElectrode(1, electrode);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 1 electrode set to " + String(electrode) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Electrode must be between 1 and 24, or 0 indicates disconnected.");
}

serialcommand(channel2Electrode) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Ch2Electrode expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int electrode = intArg(1);

  // Data Validation
  if ((electrode >= 0) and (electrode <= 24)) {
    MUX.setElectrode(2, electrode);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 2 electrode set to " + String(electrode) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Electrode must be between 1 and 24, or 0 indicates disconnected.");
}

serialcommand(channel3Electrode) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Ch3Electrode expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int electrode = intArg(1);

  // Data Validation
  if ((electrode >= 0) and (electrode <= 24)) {
    MUX.setElectrode(3, electrode);
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 3 electrode set to " + String(electrode) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Electrode must be between 1 and 24, or 0 indicates disconnected.");
}

serialcommand(channel4Electrode) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Ch4Electrode expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int electrode = intArg(1);

  // Data Validation
  if ((electrode >= 0) and (electrode <= 24)) {
    MUX.setElectrode(4, electrode);
  
    // Ripple out the new MUX values.
    MUX.sendUpdate();
    
    if (debugSerial) SerialUSB.println("SUCCESS: Channel 4 electrode set to " + String(electrode) + ".");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Electrode must be between 1 and 24, or 0 indicates disconnected.");
}

serialcommand(setHighCurrentMode) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: SetHighCurrentMode expects one argument, received " + String(numArgs()) + ".");
    return;
  }

  int value = intArg(1);
  if (value == 0) {
    MDAC0.SetLowCurrent();
    MDAC1.SetLowCurrent();
    MDAC0.UpdateCurrent();
    MDAC1.UpdateCurrent();
    if (debugSerial) SerialUSB.println("SUCCESS: Configured for low current 0-2.0mA precision mode.");
  }
  else if (value == 1) {
    MDAC0.SetHighCurrent();
    MDAC1.SetHighCurrent();
    MDAC0.UpdateCurrent();
    MDAC1.UpdateCurrent();
    if (debugSerial) SerialUSB.println("SUCCESS: Configured for high current 0-20mA mode.");
  }
  else if (debugSerial) SerialUSB.println("ERROR: High Current Mode must be set to either 1 (ON) or 0 (OFF).");
}

serialcommand(setMDAC0) {
  if (numArgs() != 2) { // Note the quirky case since it will detect mA as the first argument.
    if (debugSerial) SerialUSB.println("ERROR: Wfm0 mA expects one argument, received " + String(numArgs()-1) + ". Did you include the 'mA'?");
    return;
  }

  if (stringArg(1) != "mA") {
    if (debugSerial) SerialUSB.println("ERROR: Wfm0 mA has a typo. Did you include the 'mA'?");
    return;
  }

  float current = floatArg(2);

  // Data Validation
  if (MDAC0.GetCurrentMode()) {
    if ((current >= 0) and (current <= 20)) {
      MDAC0.SetCurrent(current);
      if (debugSerial) SerialUSB.println("SUCCESS: Waveform 0 set to " + String(current) + " mA.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: Current out of range (0-20mA).");
  }
  else {
    if ((current >= 0) and (current <= 2)) {
      MDAC0.SetCurrent(current);
      if (debugSerial) SerialUSB.println("SUCCESS: Waveform 0 set to " + String(current) + " mA.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: Current out of range (0-2.0mA).");
  }
}

serialcommand(setMDAC1) {
  if (numArgs() != 2) { // Note the quirky case since it will detect mA as the first argument.
    if (debugSerial) SerialUSB.println("ERROR: Wfm1 mA expects one argument, received " + String(numArgs()-1) + ". Did you include the 'mA'?");
    return;
  }

  if (stringArg(1) != "mA") {
    if (debugSerial) SerialUSB.println("ERROR: Wfm1 mA has a typo. Did you include the 'mA'?");
    return;
  }

  float current = floatArg(2);
  
  // Data Validation
  if (MDAC1.GetCurrentMode()) {
    if ((current >= 0) and (current <= 20)) {
      MDAC1.SetCurrent(current);
      if (debugSerial) SerialUSB.println("SUCCESS: Waveform 0 set to " + String(current) + " mA.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: Current out of range (0-20mA).");
  }
  else {
    if ((current >= 0) and (current <= 2)) {
      MDAC1.SetCurrent(current);
      if (debugSerial) SerialUSB.println("SUCCESS: Waveform 0 set to " + String(current) + " mA.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: Current out of range (0-2.0mA).");
  }
}

serialcommand(rampMDAC0) {
  if (numArgs() != 4) { // Note the quirky case since it will detect mA as the first argument.
    if (debugSerial) SerialUSB.println("ERROR: RampWfm0 mA expects three arguments, received " + String(numArgs()-1) + ". Did you include the 'mA'?");
    return;
  }

  if (stringArg(1) != "mA") {
    if (debugSerial) SerialUSB.println("ERROR: RampWfm0 mA has a typo. Did you include the 'mA'?");
    return;
  }

  float startVal = floatArg(2);
  float endVal = floatArg(3);
  float period = floatArg(4);

  if (startVal >= 0 and startVal <= 20) {
    if (endVal >= 0 and endVal <= 20) {
      if (period > 0) {
        MDAC0Ramping = true;
        MDAC0StartTime = micros();
        MDAC0.SetCurrent(startVal);
        MDAC0StartCurrent = startVal;
        MDAC0EndCurrent = endVal;
        MDAC0RampTime = period*1000;
        if (debugSerial) SerialUSB.println("SUCCESS: Waveform 0 ramping from " + String(startVal) + " mA to " + String(endVal) + " mA over " + String(period) + " ms.");
      }
      else if (debugSerial) SerialUSB.println("ERROR: Period must be greater than zero milliseconds.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: End Value must be 0-20mA.");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Start Value must be 0-20mA."); 
}

serialcommand(rampMDAC1) {
  if (numArgs() != 4) { // Note the quirky case since it will detect mA as the first argument.
    if (debugSerial) SerialUSB.println("ERROR: RampWfm1 mA expects three arguments, received " + String(numArgs()-1) + ". Did you include the 'mA'?");
    return;
  }

  if (stringArg(1) != "mA") {
    if (debugSerial) SerialUSB.println("ERROR: RampWfm1 mA has a typo. Did you include the 'mA'?");
    return;
  }

  float startVal = floatArg(2);
  float endVal = floatArg(3);
  float period = floatArg(4);

  if (startVal >= 0 and startVal <= 20) {
    if (endVal >= 0 and endVal <= 20) {
      if (period > 0) {
        MDAC1Ramping = true;
        MDAC1StartTime = micros();
        MDAC1.SetCurrent(startVal);
        MDAC1StartCurrent = startVal;
        MDAC1EndCurrent = endVal;
        MDAC1RampTime = period*1000;
        if (debugSerial) SerialUSB.println("SUCCESS: Waveform 1 ramping from " + String(startVal) + " mA to " + String(endVal) + " mA over " + String(period) + " ms.");
      }
      else if (debugSerial) SerialUSB.println("ERROR: Period must be greater than zero milliseconds.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: End Value must be 0-20mA.");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Start Value must be 0-20mA."); 
  
}

serialcommand(rampMDACs) {
  if (numArgs() != 6) { // Note the quirky case since it will detect mA as the first argument.
    if (debugSerial) SerialUSB.println("ERROR: RampWfm mA expects five arguments, received " + String(numArgs()-1) + ". Did you include the 'mA'?");
    return;
  }

  if (stringArg(1) != "mA") {
    if (debugSerial) SerialUSB.println("ERROR: RampWfm mA has a typo. Did you include the 'mA'?");
    return;
  }

  float startVal0 = floatArg(2);
  float endVal0 = floatArg(3);
  float startVal1 = floatArg(4);
  float endVal1 = floatArg(5);
  float period = floatArg(6);

  if (startVal0 >= 0 and startVal0 <= 20 and startVal1 >= 0 and startVal1 <= 20) {
    if (endVal0 >= 0 and endVal0 <= 20 and endVal1 >= 0 and endVal1 <= 20) {
      if (period > 0) {
        MDAC0Ramping = true;
        MDAC1Ramping = true;
        MDAC0StartTime = micros();
        MDAC1StartTime = MDAC0StartTime;
        MDAC0.SetCurrent(startVal0);
        MDAC1.SetCurrent(startVal1);
        MDAC0StartCurrent = startVal0;
        MDAC1StartCurrent = startVal1;
        MDAC0EndCurrent = endVal0;
        MDAC1EndCurrent = endVal1;
        MDAC0RampTime = period*1000;
        MDAC1RampTime = MDAC0RampTime;
        if (debugSerial) {
          SerialUSB.println("SUCCESS: Waveform 0 ramping from " + String(startVal0) + " mA to " + String(endVal0) + " mA over " + String(period) + " ms.");
          SerialUSB.println("SUCCESS: Waveform 1 ramping from " + String(startVal1) + " mA to " + String(endVal1) + " mA over " + String(period) + " ms.");
        }
      }
      else if (debugSerial) SerialUSB.println("ERROR: Period must be greater than zero milliseconds.");
    }
    else if (debugSerial) SerialUSB.println("ERROR: End Values must be 0-20mA.");
  }
  else if (debugSerial) SerialUSB.println("ERROR: Start Values must be 0-20mA."); 
}

serialcommand(dacWrite) {
  if (numArgs() != 2) {
    if (debugSerial) {
      SerialUSB.println("ERROR: DAC expects two arguments, received " + String(numArgs()) + ".");
    }
    return;
  }

  byte channel = intArg(1);

  if (channel != 0 and channel != 1) {
    if (debugSerial) { 
      SerialUSB.println("ERROR: Channel must be 0 or 1.");
      SerialUSB.println("Note: DAC is not backwards compatible, only the two onboard Due DACs are available.");
    }
    return;
  }
  
  float value = floatArg(2);

  if ((value < 0.55 or value > 2.75) and (value != 0)) {
    if (debugSerial) SerialUSB.println("ERROR: Due Onboard DAC can only output 0.55-2.75V (or ground).");
    return;
  }

  if (channel == 0) {
    if (value == 0) digitalWrite(DAC0, LOW);
    else analogWrite(DAC0, constrain(4096*(value-0.55)/2.2, 0, 4095));
    if (debugSerial) SerialUSB.println("SUCCESS: DAC0 set to " + String(value) + " V.");
  }
  else if (channel == 1) {
    if (value == 0) digitalWrite(DAC1, LOW);
    else analogWrite(DAC1, constrain(4096*(value-0.55)/2.2, 0, 4095));
    if (debugSerial) SerialUSB.println("SUCCESS: DAC1 set to " + String(value) + " V.");
  }
  else {
    if (debugSerial) SerialUSB.println("ERROR: Due Onboard DAC can only output 0.55-2.75V.");    
  }
}

serialcommand(wssWrite) {
  if (numArgs() != 2) {
    if (debugSerial) SerialUSB.println("ERROR: WSS expects two arguments, received " + String(numArgs()) + ".");
    return;
  }

  byte channel = intArg(1);

  if (channel != 0 and channel != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Channel must be 0 or 1.");
    return;
  }

  String source = stringArg(2);

  byte code;
  if (source == "SIN0") code = SRC_SINE0;
  else if (source == "SIN1") code = SRC_SINE1;
  else if (source == "DAC0") code = SRC_DAC0;
  else if (source == "DAC1") code = SRC_DAC1;
  else if (source == "NONE") code = SRC_NONE;
  else {
    if (debugSerial) SerialUSB.println("ERROR: Source may be SIN0, SIN1, DAC0, DAC1, or NONE.");
    return;
  }

  if (channel == 0) {
    WSS0.setSource(code);
    if (debugSerial) SerialUSB.println("SUCCESS: WSS0 set to " + source);
  }
  else if (channel == 1) {
    WSS1.setSource(code);
    if (debugSerial) SerialUSB.println("SUCCESS: WSS1 set to " + source);
  }
  else {
    if (debugSerial) SerialUSB.println("ERROR: Channel must be 0 or 1.");
  }
}

serialcommand(readDiagnosticSignals) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: ReadDiagnosticSignals expects no arguments, received " + String(numArgs()) + ".");
    return;
  }
  SerialUSB.println("+50V: " + String((ExtADC.getValue(4)+ExtADC.getVCM())*21,10));
  SerialUSB.println("-50V: " + String((ExtADC.getValue(5)+ExtADC.getVCM())*21 - 20*3.3,10));
  SerialUSB.println("Battery: " + String(ExtADC.getDUEScaler()*(float)analogRead(A10) * 111.0/11.0,10));
  SerialUSB.println("1.65V: " + String(ExtADC.getValue(6)+ExtADC.getVCM(),10));
}

serialcommand(adcRead) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: ADCRead expects one argument (delay between samples in us), received " + String(numArgs()) + ".");
    return;
  }

  int ADCdelay = intArg(1);

  // Data validation.
  if (ADCdelay <= 0) {
    if (debugSerial) SerialUSB.println("ERROR: Delay time must be greater than zero.");
    return;
  }

  // Collect and then stream data.
  unsigned int analogvalue0, analogvalue1, analogvalue4, analogvalue5;
  uint8_t analog[32000];
  noInterrupts();
  for (int i = 0; i < 32000; i = i + 8) {
    analogvalue0 = analogRead(0);
    analogvalue1 = analogRead(1);
    analogvalue4 = analogRead(4);
    analogvalue5 = analogRead(5);
    analog[i] = (analogvalue0 >> 8) & 0xFF;
    analog[i + 1] = analogvalue0 & 0xFF;
    analog[i + 2] = (analogvalue1 >> 8) & 0xFF;
    analog[i + 3] = analogvalue1 & 0xFF;
    analog[i + 4] = (analogvalue4 >> 8) & 0xFF;
    analog[i + 5] = analogvalue4 & 0xFF;
    analog[i + 6] = (analogvalue5 >> 8) & 0xFF;
    analog[i + 7] = analogvalue5 & 0xFF;
    delayMicroseconds(ADCdelay);
  }
  interrupts();
  
  SerialUSB.write(analog, 32000);
}

serialcommand(setDCMode) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetDCMode expects no arguments, received " + String(numArgs()) + ".");
    return;
  }
  MUX.DCMode();
  if (debugSerial) SerialUSB.println("SUCCESS: DC coupled mode enabled.");
}

serialcommand(setACMode) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetACMode expects no arguments, received " + String(numArgs()) + ".");
    return;
  }
  MUX.ACMode();
  if (debugSerial) SerialUSB.println("SUCCESS: AC coupled mode enabled.");
}

serialcommand(setModeRun) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetModeRun expects no arguments, received " + String(numArgs()) + ".");
    return;
  }
  SetModeRun();
  if (debugSerial) SerialUSB.println("SUCCESS: Current active, MUX Electrodes active.");
}

void SetModeRun(void) {
  MDAC0.Enable();
  MDAC1.Enable();
  MUX.EnableOutput();
  stimulationstate = RUN;
}

serialcommand(setModeDebug) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetModeDebug expects no arguments, received " + String(numArgs()) + ".");
    return;
  }
  SetModeDebug();
  if (debugSerial) SerialUSB.println("SUCCESS: Current active, MUX Electrodes disconnected, 10k attached.");
}

void SetModeDebug(void) {
  MDAC0.Enable();
  MDAC1.Enable();
  MUX.DisableOutput();
  stimulationstate = DEBUG;
}

serialcommand(setModeSafe) {
  if (numArgs() != 0) {
    if (debugSerial) SerialUSB.println("ERROR: SetModeSafe expects no arguments, received " + String(numArgs()) + ".");
    return;
  }
  SetModeSafe();
  if (debugSerial) SerialUSB.println("SUCCESS: Current disabled, MUX Electrodes disconnected, 10k attached..");
}

void SetModeSafe(void) {
  MDAC0.Disable();
  MDAC1.Disable();
  MUX.DisableOutput();
  stimulationstate = SAFE;
}

serialcommand(stimulate) {
  if (numArgs() != 1) {
    if (debugSerial) SerialUSB.println("ERROR: Stimulate expects one argument (run period in milliseconds), received " + String(numArgs()) + ".");
    return;
  }

  unsigned int endtime = intArg(1);

  if (endtime <= 0) {
    if (debugSerial) SerialUSB.println("ERROR: Stimulation time must be greater than zero (milliseconds).");
    return;
  }

  if (debugSerial) SerialUSB.println("SUCCESS: Stimulation started for " + String(endtime) + " ms.");
  endtime += millis();
  
  enum stimulationstate stimulationstatesave = stimulationstate;
  SetModeRun();
  while (millis() < endtime);
  if (stimulationstatesave == RUN); // Do nothing, it's already running.
  else if (stimulationstatesave == DEBUG) SetModeDebug();
  else if (stimulationstatesave == SAFE) SetModeSafe();
  else {
    if (debugSerial) SerialUSB.println("ERROR: Saved state broke somehow and was not RUN, DEBUG, or SAFE. Check Code for bugs, this shouldn't happen.");
    return;
  }

  if (debugSerial) SerialUSB.println("SUCCESS: Stimulation Completed.");
}
